(*
 * coqweb - A WEB-like tool for Coq
 * Copyright (C) 2000 Jean-Christophe FILLIATRE
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * See the GNU General Public License version 2 for more details
 * (enclosed in the file GPL).
 *)

(*i $Id: doclexer.mll,v 1.4 2001/06/07 07:04:50 filliatr Exp $ i*)

{
  open Printf
  open Lexing
  open Output
  open Web

(*s Global variables and functions used by the lexer. *)

(* [skip_header] tells whether option \verb|--header| has been
   selected by the user. *)

  let skip_header = ref true

(* For displaying error message if any, [current_file] records the
   name of the file currently read, and [comment_or_string_start]
   records the starting position of the comment or the string
   currently being read. *)

  let current_file = ref ""
  let comment_or_string_start = ref 0

(* [web_style] records if web sections were used anywhere in any file. *)

  let web_style = ref false

(*s Global variables for temporarily recording data, for building
    the [Web.file] structure. *)
				  
  let parbuf = Buffer.create 8192
  let ignoring = ref false

  let push_char c = 
    if not !ignoring then Buffer.add_char parbuf c
  let push_first_char lexbuf = 
    if not !ignoring then Buffer.add_char parbuf (lexeme_char lexbuf 0)
  let push_string s = 
    if not !ignoring then Buffer.add_string parbuf s

  let parlist = ref ([] : paragraph list)
  let code_beg = ref 0

  let push_code () =
    if Buffer.length parbuf > 0 then begin
      parlist := (Code (!code_beg, Buffer.contents parbuf)) :: !parlist;
      Buffer.clear parbuf
    end

  let push_doc () =
    if Buffer.length parbuf > 0 then begin
      parlist := (Documentation (Buffer.contents parbuf)) :: !parlist;
      Buffer.clear parbuf
    end

  let seclist = ref ([] : raw_section list)
  let section_beg = ref 0

  let push_section () =
    if !parlist <> [] then begin
      let s = { sec_contents = List.rev !parlist; sec_beg = !section_beg } in
      seclist := s :: !seclist;
      parlist := []
    end

  let reset_lexer f =
    current_file := f;
    comment_or_string_start := 0;
    section_beg := 0;
    code_beg := 0;
    parlist := [];
    seclist := []

  let backtrack lexbuf = 
    (*i
      eprintf "backtrack to %d\n" (lexbuf.lex_abs_pos + lexbuf.lex_start_pos);
    i*)
    lexbuf.lex_curr_pos <- lexbuf.lex_start_pos

}

(*s Shortcuts for regular expressions. *)

let space = [' ' '\t']
let space_or_nl = [' ' '\t' '\n']
let character = 
  "'" ( [^ '\\' '\''] | '\\' ['\\' '\'' 'n' 't' 'b' 'r'] 
      | '\\' ['0'-'9'] ['0'-'9'] ['0'-'9'] ) "'"
let up_to_end_of_comment =
  [^ '*']* '*' (([^ '*' ')'] [^ '*']* '*') | '*')* ')'

(*s Entry point to skip the headers. Returns when headers are skipped. *)
rule header = parse
  | "(*"
      { comment_or_string_start := lexeme_start lexbuf;
	skip_comment lexbuf;
	skip_spaces_until_nl lexbuf; 
	header lexbuf }
  | "\n" 
      { () }
  | space+ 
      { header lexbuf }
  | _      
      { backtrack lexbuf }
  | eof    
      { () }

(* To skip a comment (used by [header]). *)
and skip_comment = parse
  | "(*" 
      { skip_comment lexbuf; skip_comment lexbuf }
  | "*)" 
      { () }
  | eof  
      { eprintf "File \"%s\", character %d\n" 
	  !current_file !comment_or_string_start;
	eprintf "Unterminated Coq comment\n";
	exit 1 }
  | _ 
      { skip_comment lexbuf }

(*s Recognizes a complete Coq module, after header has been
   skipped. After calling that entry, the whole text read is in
   [seclist]. *)

and vernac = parse
  | _
      { backtrack lexbuf;
	paragraph lexbuf;
	vernac lexbuf }
  | eof 
      { push_section () }
      
(* Recognizes a paragraph of Coq code or documentation. After calling
   that entry, the paragraph has been added to [parlist]. *)

and paragraph = parse
  | space_or_nl* '\n'
      { paragraph lexbuf }
  | "(*" '*'* "*)" space* '\n'
      { paragraph lexbuf }
  | "(*"    
      { comment_or_string_start := lexeme_start lexbuf;
	start_of_documentation lexbuf; 
	push_doc () }
  | space* ("(*c" | _)
      { code_beg := lexeme_start lexbuf;
	backtrack lexbuf;
	coq_subparagraph lexbuf;
	push_code() }
  | eof 
      { () }
      
(*s At the beginning of the documentation part, just after the
  \verb|"(*"|. If the first character is ['s'], then a new section is
  started. After calling that entry, the [parbuf] buffer contains the
  doc read. *)


and start_of_documentation = parse
  | space_or_nl+   
      { in_documentation lexbuf }
  | 's' space_or_nl*
      { web_style := true; push_section (); 
	section_beg := lexeme_start lexbuf;
	in_documentation lexbuf }
  | 'p' up_to_end_of_comment
      { let s = lexeme lexbuf in
	push_in_preamble (String.sub s 1 (String.length s - 3)) }
  | 'i' 
      { ignore lexbuf }
  | _
      { backtrack lexbuf;
	in_documentation lexbuf }
  | eof 
      { in_documentation lexbuf }

(*s Inside the documentation part, anywhere after the "(*". After
  calling that entry, the [parbuf] buffer contains the doc read. *)

and in_documentation = parse 
  | "(*" 
      { push_string "(*";
	in_documentation lexbuf; 
	push_string "*)";
	in_documentation lexbuf }
  | "*)" 
      { () }
  | '\n' " * "
      { push_char '\n'; in_documentation lexbuf }
  | '"'  
      { push_char '"'; in_string lexbuf; in_documentation lexbuf }
  | character    
      { push_string (lexeme lexbuf); in_documentation lexbuf }
  | _    
      { push_first_char lexbuf; in_documentation lexbuf }
  | eof  
      { eprintf "File \"%s\", character %d\n" 
	  !current_file !comment_or_string_start;
	eprintf "Unterminated Coq comment\n";
	exit 1  }

(*s Recognizes a subparagraph of Coq code. After calling that entry, 
  the [parbuf] buffer contains the code read. *)

and coq_subparagraph = parse
  | space* '\n' space* '\n' 
      { backtrack lexbuf }
  | eof  
      { () }
  | "(*" | "(*c"
      { comment_or_string_start := lexeme_start lexbuf;
	push_string "(*";
	comment lexbuf; 
	coq_subparagraph lexbuf }
  | "(*i"
      { comment_or_string_start := lexeme_start lexbuf;
	ignore lexbuf; coq_subparagraph lexbuf }
  | '"'  
      { comment_or_string_start := lexeme_start lexbuf;
        push_char '"'; in_string lexbuf; coq_subparagraph lexbuf }
  | character
      { push_string (lexeme lexbuf); coq_subparagraph lexbuf }
  | _ 
      { push_first_char lexbuf; coq_subparagraph lexbuf }

(*s To skip spaces until a newline. *)
and skip_spaces_until_nl = parse
  | space* '\n'? { () }
  | eof  { () }
  | _    { backtrack lexbuf }


(*s To read a comment inside a piece of code. *)
and comment = parse
  | "(*" | "(*c"
      { push_string "(*"; comment lexbuf; comment lexbuf }
  | "*)" 
      { push_string "*)"  }
  | eof  
      { eprintf "File \"%s\", character %d\n" 
	  !current_file !comment_or_string_start;
	eprintf "Unterminated Coq comment\n";
	exit 1 }
  | _  
      { push_first_char lexbuf; comment lexbuf }

(*s Ignored parts, between "(*i" and "i*)". Note that such comments
    are not nested. *)
and ignore = parse
  | "i*)" 
      { skip_spaces_until_nl lexbuf }
  | eof   
      { eprintf "File \"%s\", character %d\n" 
	  !current_file !comment_or_string_start;
	eprintf "Unterminated coqweb comment\n"; 
	exit 1 }
  | _   
      { ignore lexbuf }

(*s Strings in code. *)
and in_string = parse
  | '"'      
      { push_char '"' }
  | '\\' ['\\' '"' 'n' 't' 'b' 'r'] 
      { push_string (lexeme lexbuf); in_string lexbuf }
  | eof      
      { eprintf "File \"%s\", character %d\n" 
	  !current_file !comment_or_string_start;
	eprintf "Unterminated Coq string\n"; 
	exit 1 }
  | _    
      { push_first_char lexbuf; in_string lexbuf }

{

(*s \textbf{Coq files.} *)

type coq_file = { coq_filename : string; coq_module : string }

let module_name f = String.capitalize (Filename.basename f)

let make_coq_file f = 
  { coq_filename = f;
    coq_module = module_name (Filename.chop_extension f) }

type file_type =
  | File_vernac  of coq_file
  | File_other of string

(*s \textbf{Reading Coq files.} *)

let raw_read_file header entry f =
  reset_lexer f;
  let c = open_in f in
  let buf = Lexing.from_channel c in
  if !skip_header then header buf;
  entry buf;
  close_in c;
  List.rev !seclist

let read header entry m =
  { content_file = m.coq_filename; 
    content_name = m.coq_module;
    content_contents = raw_read_file header entry m.coq_filename; }

let read_one_file = function
  | File_vernac m -> Vernac (read header vernac m)
  | File_other f -> Other f

}

